from datetime import datetime, timedelta

import pytest

from k8s.config import FileTokenSource


# pylint: disable=R0201
class TestFileTokenSource(object):

    @pytest.fixture
    def token_file(self, tmpdir_factory):
        fpath = tmpdir_factory.mktemp(self.__class__.__name__).join('token_file')
        try:
            yield fpath
        finally:
            if fpath.exists():
                fpath.remove()

    def test_token_read(self, token_file):
        token = "secret token"  # nosec
        token_file.write(token)
        token_source = FileTokenSource(token_file=str(token_file))

        assert token_source.token() == token

    def test_token_re_read_after_expiry(self, token_file):
        initial_token = "secret token 1"  # nosec
        token_file.write(initial_token)
        initial_time = datetime(2021, 1, 1, 10, 10)

        token_source = FileTokenSource(token_file=str(token_file), now_func=lambda: initial_time)

        assert token_source.token(now_func=lambda: initial_time) == initial_token

        updated_token = "secret token 2"  # nosec
        token_file.write(updated_token)

        assert token_source.token(now_func=lambda: initial_time + timedelta(seconds=30)) == initial_token
        assert token_source.token(now_func=lambda: initial_time + timedelta(minutes=2)) == updated_token

    def test_token_raise_if_file_does_not_exist(self, token_file):
        with pytest.raises(IOError):
            FileTokenSource(token_file=str(token_file))



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .common import ObjectMeta
from .pod import PodTemplateSpec
from ..base import Model
from ..fields import Field


class LabelSelector(Model):
    matchLabels = Field(dict)


class JobSpec(Model):
    template = Field(PodTemplateSpec)
    backoffLimit = Field(int)
    activeDeadlineSeconds = Field(int)
    completions = Field(int)
    manualSelector = Field(bool)
    parallelism = Field(int)
    selector = Field(LabelSelector)


class JobTemplateSpec(Model):
    metadata = Field(ObjectMeta)
    spec = Field(JobSpec)


class Job(Model):
    class Meta:
        list_url = "/apis/batch/v1/jobs"
        url_template = "/apis/batch/v1/namespaces/{namespace}/jobs"

    metadata = Field(ObjectMeta)
    spec = Field(JobSpec)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import pytest

from k8s.client import NotFound
from k8s.models.common import ObjectMeta, LabelSelector
from k8s.models.policy_v1_pod_disruption_budget import PodDisruptionBudget, PodDisruptionBudgetSpec

NAME = "my-name"
NAMESPACE = "my-namespace"
PDB_URI = PodDisruptionBudget._meta.url_template.format(name="", namespace=NAMESPACE)


@pytest.mark.usefixtures("logger", "k8s_config")
class TestPodDisruptionBudget(object):
    def test_create_blank_pdb(self):
        pdb = _create_pdb()
        assert pdb.metadata.name == NAME
        assert pdb.as_dict()["metadata"]["name"] == NAME

    def test_pdb_created_if_not_exists(self, post, api_get):
        api_get.side_effect = NotFound()
        pdb = _create_pdb()
        call_params = pdb.as_dict()
        post.return_value.json.return_value = call_params

        assert pdb._new
        pdb.save()
        assert not pdb._new
        pytest.helpers.assert_any_call(post, PDB_URI, call_params)

    def test_get_or_create_pdb_not_new(self, put, get):
        mock_response = mock.Mock()
        mock_response.json.return_value = {
            'apiVersion': 'policy/v1',
            'kind': 'PodDisruptionBudget',
            'metadata': {
                'creationTimestamp': '2017-10-03T10:36:20Z',
                'labels': {
                    'app': 'my-name', 'test': 'true'
                },
                'name': 'my-name',
                'namespace': 'my-namespace',
                'resourceVersion': '852',
                'uid': 'b1e35ab5-a826-11e7-ba76-0800273598c9'
            },
            'spec': {
                'minAvailable': 1,
                'selector': {
                    'matchLabels': {
                        'app': 'my-name',
                    },
                },
            },
            'status': {
                'currentHealthy': 1,
                'desiredHealthy': 1,
                'disruptedPods': {},
                'expectedPods': 1,
                'observedGeneration': 1
            }
        }
        get.return_value = mock_response
        pdb = PodDisruptionBudget.get(name=NAME, namespace=NAMESPACE)
        assert not pdb._new
        assert pdb.metadata.name == NAME
        assert pdb.spec.minAvailable == 1
        assert pdb.spec.selector.matchLabels["app"] == NAME
        call_params = pdb.as_dict()
        put.return_value.json.return_value = call_params

        pdb.save()
        pytest.helpers.assert_any_call(put, PDB_URI + NAME, call_params)

    def test_pdb_deleted(self, delete):
        PodDisruptionBudget.delete(NAME, NAMESPACE)

        # call delete with service_name
        pytest.helpers.assert_any_call(delete, PDB_URI + NAME)


def _create_pdb():
    object_meta = ObjectMeta(name=NAME, namespace=NAMESPACE, labels={"test": "true", "app": NAME})
    pdb_spec = PodDisruptionBudgetSpec(minAvailable=1, selector=LabelSelector(matchLabels={"app": NAME}))
    first = PodDisruptionBudget(metadata=object_meta, spec=pdb_spec)
    return first



#!/usr/bin/env python
# -*- coding: utf-8
import base64
import copy

import mock
import pytest

from k8s.client import NotFound
from k8s.models.common import ObjectMeta
from k8s.models.secret import Secret

# Copyright 2017-2022 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NAME = "my-name"
NAMESPACE = "my-namespace"
STRING_DATA = {"boo": "baz"}


# pylint: disable=R0201
@pytest.mark.usefixtures("k8s_config")
class TestSecret(object):
    def test_created_if_not_exists(self, post, api_get):
        api_get.side_effect = NotFound()
        secret = _create_default_secret()
        call_params = secret.as_dict()
        ret_val = _convert_string_data(call_params)
        post.return_value.json.return_value = ret_val

        assert secret._new
        secret.save()
        assert not secret._new

        pytest.helpers.assert_any_call(post, _uri(NAMESPACE), call_params)

    def test_updated_if_exists(self, get, put):
        mock_response = _create_mock_response()
        get.return_value = mock_response
        secret = _create_default_secret()

        from_api = Secret.get_or_create(metadata=secret.metadata, data=secret.data, type=secret.type)
        assert not from_api._new
        assert from_api.data == {"token": "NWVtaXRq"}

        from_api.data = {"boo": "baz"}
        call_params = from_api.as_dict()
        put.return_value.json.return_value = call_params

        from_api.save()
        pytest.helpers.assert_any_call(put, _uri(NAMESPACE, NAME), call_params)

    def test_deleted(self, delete):
        Secret.delete(NAME, namespace=NAMESPACE)
        pytest.helpers.assert_any_call(delete, _uri(NAMESPACE, NAME))

    def test_string_data(self):
        secret = _create_default_secret()
        serialized = secret.as_dict()
        assert serialized["stringData"] == STRING_DATA


def _create_mock_response():
    mock_response = mock.Mock()
    mock_response.json.return_value = {
        "apiVersion": "v1",
        "kind": "Secret",
        "metadata": {
            "creationTimestamp": "2022-05-01T13:37:00Z",
            "generation": 1,
            "labels": {
                "test": "true"
            },
            "name": NAME,
            "namespace": NAMESPACE,
            "resourceVersion": "178",
            "selfLink": _uri(NAMESPACE, NAME),
            "uid": "51ae75dc-c3fc-4b08-842f-1d546b88ae98"
        },
        "data": {
            "token": "NWVtaXRq",
        },
        "type": "kubernetes.io/tls",
    }
    return mock_response


def _create_default_secret():
    object_meta = ObjectMeta(name=NAME, namespace=NAMESPACE, labels={"test": "true"})
    data = {"token": "NWVtaXRq"}
    object_type = "kubernetes.io/tls"
    string_data = STRING_DATA
    secret = Secret(metadata=object_meta, data=data, type=object_type, stringData=string_data)
    return secret


def _uri(namespace, name=""):
    return "/api/v1/namespaces/{namespace}/secrets/{name}".format(name=name, namespace=namespace)


def _convert_string_data(call_params):
    converted = copy.deepcopy(call_params)
    string_data = converted.get("stringData")
    if string_data:
        del converted["stringData"]
        for key, value in string_data.items():
            converted["data"][key] = base64.b64encode(value.encode("utf-8"))
    return converted



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2022 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
import pytest

from k8s.client import NotFound
from k8s.models.common import ObjectMeta
from k8s.models.certificate import Certificate, CertificateSpec, CertificateStatus, \
    IssuerReference, CertificateCondition

NAME = "my-name"
NAMESPACE = "my-namespace"
REF_NAME = "example_tls"


# pylint: disable=R0201
@pytest.mark.usefixtures("k8s_config")
class TestCertificate(object):
    def test_created_if_not_exists(self, post, api_get):
        api_get.side_effect = NotFound()
        cert = _create_default_certificate()
        call_params = cert.as_dict()
        post.return_value.json.return_value = call_params

        assert cert._new
        cert.save()
        assert not cert._new

        pytest.helpers.assert_any_call(post, _uri(NAMESPACE), call_params)

    def test_updated_if_exists(self, get, put):
        mock_response = _create_mock_response()
        get.return_value = mock_response
        cert = _create_default_certificate()

        from_api = Certificate.get_or_create(metadata=cert.metadata, spec=cert.spec, status=cert.status)
        assert not from_api._new
        assert from_api.spec.secretName == REF_NAME

        from_api.data = {"boo": "baz"}
        call_params = from_api.as_dict()
        put.return_value.json.return_value = call_params

        from_api.save()
        pytest.helpers.assert_any_call(put, _uri(NAMESPACE, NAME), call_params)

    def test_deleted(self, delete):
        Certificate.delete(NAME, namespace=NAMESPACE)
        pytest.helpers.assert_any_call(delete, _uri(NAMESPACE, NAME))


def _create_mock_response():
    mock_response = mock.Mock()
    mock_response.json.return_value = {
        "apiVersion": "cert-manager.io/v1",
        "kind": "Certificate",
        "metadata": {
            "creationTimestamp": "2022-05-01T13:37:00Z",
            "generation": 1,
            "labels": {
                "test": "true"
            },
            "name": NAME,
            "namespace": NAMESPACE,
            "resourceVersion": "178",
            "selfLink": _uri(NAMESPACE, NAME),
            "uid": "51ae75dc-c3fc-4b08-842f-1d546b88ae98"
        },
        "spec": {
            "secretName": REF_NAME,
            "issuerRef": {
                "name": "test",
                "kind": "Issuer"
            }
        },
        "status": {
            "conditions": [
                {
                    "type": "Ready",
                    "status": "True"
                }
            ]
        },
    }
    return mock_response


def _create_default_certificate():
    object_meta = ObjectMeta(name=NAME, namespace=NAMESPACE, labels={"test": "true"})
    spec = CertificateSpec(secretName=REF_NAME, issuerRef=IssuerReference(name="test", kind="Issuer"))
    status = CertificateStatus(conditions=[CertificateCondition(type="Ready", status="True")])
    cert = Certificate(metadata=object_meta, spec=spec, status=status)
    return cert


def _uri(namespace, name=""):
    uri = "/apis/cert-manager.io/v1/namespaces/{namespace}/certificates/{name}"
    return uri.format(name=name, namespace=namespace)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .common import ObjectMeta, LabelSelector, Condition
from ..base import Model
from ..fields import Field, ListField


class PodDisruptionBudgetSpec(Model):
    minAvailable = Field(str, alt_type=int)
    maxUnavailable = Field(str, alt_type=int)
    selector = Field(LabelSelector)
    unhealthyPodEvictionPolicy = Field(str)


class PodDisruptionBudgetStatus(Model):
    conditions = ListField(Condition)
    currentHealthy = Field(int)
    desiredHealthy = Field(int)
    disruptionsAllowed = Field(int)
    disruptedPods = Field(dict)
    expectedPods = Field(int)
    observedGeneration = Field(int)


class PodDisruptionBudget(Model):
    class Meta:
        list_url = "/apis/policy/v1/poddisruptionbudgets"
        url_template = "/apis/policy/v1/namespaces/{namespace}/poddisruptionbudgets/{name}"

    metadata = Field(ObjectMeta)
    spec = Field(PodDisruptionBudgetSpec)
    status = Field(PodDisruptionBudgetStatus)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2022 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from .common import ObjectMeta
from ..base import Model
from ..fields import Field, ListField


class IssuerReference(Model):
    name = Field(str)
    kind = Field(str)
    group = Field(str)


class CertificateSpec(Model):
    secretName = Field(str)
    issuerRef = Field(IssuerReference)


class CertificateCondition(Model):
    type = Field(str)
    status = Field(str)


class CertificateStatus(Model):
    conditions = ListField(CertificateCondition)
    notAfter = Field(datetime.datetime)


class Certificate(Model):
    class Meta:
        list_url = "/apis/cert-manager.io/v1/certificates"
        url_template = "/apis/cert-manager.io/v1/namespaces/{namespace}/certificates/{name}"

    metadata = Field(ObjectMeta)
    spec = Field(CertificateSpec)
    status = Field(CertificateStatus)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from setuptools import setup, find_packages

GENERIC_REQ = [
    "requests==2.31.0",
    "pyrfc3339==1.1",
    "cachetools==3.1.1",  # Newer versions require Python 3
]

CODE_QUALITY_REQ = [
    'prospector==1.2.0',  # Newer versions require Python 3
]

TESTS_REQ = [
    'tox==3.24.5',
    'mock==3.0.5',  # Newer versions require Python 3
    "pytest-sugar==0.9.4",
    "pytest-html==1.22.0",  # Newer versions require Python 3
    "pytest-cov==2.7.1",
    "pytest-helpers-namespace==2019.1.8",
    'pytest==4.6.11',  # Newer versions require Python 3
]


def _generate_description():
    description = [_read("README.rst")]
    changelog_file = os.getenv("CHANGELOG_FILE")
    if changelog_file:
        description.append(_read(changelog_file))
    return "\n".join(description)


def _get_license_name():
    with open(os.path.join(os.path.dirname(__file__), "LICENSE")) as f:
        for line in f:
            if line.strip():
                return line.strip()


def _read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return f.read()


def main():
    setup(
        name="k8s",
        author="FiaaS developers",
        author_email="fiaas@googlegroups.com",
        use_scm_version=True,
        packages=find_packages(exclude=("tests",)),
        zip_safe=True,
        include_package_data=True,

        # Requirements
        install_requires=GENERIC_REQ,
        setup_requires=['pytest-runner', 'wheel', 'setuptools_scm'],
        extras_require={
            "dev": TESTS_REQ + CODE_QUALITY_REQ,
            "codacy": ["codacy-coverage"],
            "docs": ["Sphinx>=1.6.3"]
        },
        tests_require=TESTS_REQ,
        # Metadata
        description="Python client library for the Kubernetes API",
        long_description=_generate_description(),
        url="https://github.com/fiaas/k8s",
        license=_get_license_name(),
        keywords="kubernetes fiaas",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Internet",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: System :: Clustering",
            "Topic :: System :: Distributed Computing",
        ]
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=R0201

from datetime import datetime

import mock
import pytest
import pytz

from k8s import config
from k8s.base import Model, SelfModel
from k8s.fields import Field, JSONField, ListField, OnceField, ReadOnlyField, RequiredField, WriteOnlyField
from k8s.models.common import ObjectMeta

NAME = "my-model-test"


class ModelTest(Model):
    class Meta:
        url_template = "/apis/v1/modeltests/{name}"

    metadata = Field(ObjectMeta)
    field = Field(int)
    list_field = ListField(int)
    once_field = OnceField(int)
    read_only_field = ReadOnlyField(int)
    write_only_field_int = WriteOnlyField(int)
    write_only_field_dict = WriteOnlyField(dict)
    alt_type_field = Field(int, alt_type=str)
    dict_field = Field(dict)
    _exec = Field(int)
    time_field = Field(datetime)
    json_field = JSONField()
    self_field = Field(SelfModel)


class TestFields(object):
    @pytest.fixture(autouse=True)
    def set_config_debug(self, monkeypatch):
        monkeypatch.setattr(config, "debug", True)

    @pytest.mark.parametrize("field_name,initial_value,other_value", [
        ("field", 1, 2),
        ("list_field", [1], [1, 2]),
        ("once_field", 1, 2),
        ("_exec", 1, 2),
        ("json_field", 1, [1, 2]),
    ])
    def test_field_new(self, field_name, initial_value, other_value):
        kwargs = {"new": True, field_name: initial_value}
        model = ModelTest(**kwargs)
        assert getattr(model, field_name) == initial_value
        setattr(model, field_name, other_value)
        assert getattr(model, field_name) == other_value

    @pytest.mark.parametrize("field_name,initial_value,get_value", [
        ("write_only_field_int", 1, 1),
        ("write_only_field_dict", {1: 1}, {1: 1}),
    ])
    def test_write_only_field(self, field_name, initial_value, get_value):
        kwargs = {"new": True, field_name: initial_value, "metadata": ObjectMeta(name="name")}
        model = ModelTest(**kwargs)
        assert getattr(model, field_name) == get_value
        data = model.as_dict()
        assert data[field_name] == initial_value
        from_api = ModelTest.from_dict(data)
        assert getattr(from_api, field_name) is None

    @pytest.mark.parametrize("field_name,initial_value,other_value", [
        ("field", 1, 2),
        ("list_field", [1], [1, 2]),
    ])
    def test_field_old(self, field_name, initial_value, other_value):
        model = ModelTest.from_dict({field_name: initial_value})
        assert getattr(model, field_name) == initial_value
        setattr(model, field_name, other_value)
        assert getattr(model, field_name) == other_value

    def test_once_field_old(self):
        model = ModelTest.from_dict({"once_field": 1})
        assert model.once_field == 1
        model.once_field = 2
        assert model.once_field == 1

    def test_exec_field_old(self):
        model = ModelTest.from_dict({"exec": 1})
        assert model._exec == 1
        model._exec = 2
        assert model._exec == 2
        assert model.as_dict()[u"exec"] == 2

    def test_read_only_field_new(self):
        model = ModelTest(new=True, read_only_field=1)
        assert model.read_only_field is None
        model.read_only_field = 2
        assert model.read_only_field is None

    def test_read_only_field_old(self):
        model = ModelTest.from_dict({"read_only_field": 1})
        assert model.read_only_field == 1
        model.read_only_field = 2
        assert model.read_only_field == 1

    @pytest.mark.parametrize("value,modifier", [
        (1, lambda x: x + 1),
        (u"string", lambda x: x.upper())
    ])
    def test_alt_type_field(self, value, modifier):
        model = ModelTest.from_dict({"alt_type_field": value})
        assert model.alt_type_field == value
        assert model.as_dict()[u"alt_type_field"] == value
        model.alt_type_field = modifier(value)
        assert model.alt_type_field == modifier(value)

    @pytest.mark.parametrize("input,dt", [
        ("2009-01-01T17:59:59Z", datetime(2009, 1, 1, 17, 59, 59, tzinfo=pytz.UTC)),
        ("2009-01-01T17:59:59+01:00", datetime(2009, 1, 1, 16, 59, 59, tzinfo=pytz.UTC)),
    ])
    def test_time_field_from_dict(self, input, dt):
        model = ModelTest.from_dict({"time_field": input})
        assert isinstance(model.time_field, datetime)
        assert model.time_field == dt

    def test_time_field_as_dict(self):
        model = ModelTest(time_field=datetime(2009, 1, 1, 17, 59, 59, tzinfo=pytz.UTC))
        d = model.as_dict()
        assert d["time_field"] == "2009-01-01T17:59:59Z"

    @pytest.mark.parametrize("value,is_valid", [
        (None, True),
        (1, True),
        (1.1, True),
        ("string", True),
        ([1, 2], True),
        ({"key": "value"}, True),
        (ModelTest(), False),
        ([1, None], False),
    ])
    def test_json_field_validation(self, value, is_valid):
        model = ModelTest()
        if is_valid:
            setattr(model, "json_field", value)
            assert model.json_field == value
        else:
            with pytest.raises(TypeError):
                setattr(model, "json_field", value)
            assert model.json_field is None


class RequiredFieldTest(Model):
    required_field = RequiredField(int)
    field = Field(int, 100)


@pytest.mark.usefixtures("logger")
class TestRequiredField(object):
    @pytest.mark.parametrize("kwargs", [
        {"required_field": 1, "field": 2},
        {"required_field": 1},
    ])
    def test_create_with_fields(self, kwargs):
        instance = RequiredFieldTest(new=True, **kwargs)
        for key, value in kwargs.items():
            assert getattr(instance, key) == value

    def test_create_fails_when_field_missing(self):
        with pytest.raises(TypeError):
            RequiredFieldTest(new=True, field=1)


class TestSelfField(object):
    def test_create_from_dict(self):
        model = ModelTest.from_dict({"self_field": {"exec": 1}})
        assert getattr(model, "self_field") == ModelTest(_exec=1)

    def test_get_or_create_merge(self, get):
        get.return_value = _create_mock_response()
        object_meta = ObjectMeta(name=NAME, labels={"test": "true"})
        model = ModelTest.get_or_create(metadata=object_meta, self_field=ModelTest(_exec=1))
        assert getattr(model, "self_field") == ModelTest.from_dict({"read_only_field": 1, "exec": 1})


def _create_mock_response():
    mock_response = mock.Mock()
    mock_response.json.return_value = {
        "apiVersion": "v1",
        "kind": "ModelTest",
        "metadata": {
            "creationTimestamp": "2019-11-23T13:43:42Z",
            "generation": 7,
            "labels": {
                "test": "true"
            },
            "name": NAME,
            "resourceVersion": "96758807",
            "selfLink": _uri(NAME),
            "uid": "d8f1ba26-b182-11e6-a364-fa163ea2a9c4"
        },
        "self_field": {
            "read_only_field": 1
        }
    }
    return mock_response


def _uri(name=""):
    return "/apis/v1/modeltests/{name}".format(name=name)
